from recurfaces import Recurface
from objectextensions import Extendable
from pygame import Surface

from abc import ABC
from typing import Optional, Tuple, Any


class Renderable(Extendable, Recurface, ABC):
    """
    A Renderable object is any object that has a place in the rendering hierarchy.

    Explanation of additional constructor param requirements:
    - game: A reference to the Game instance allows all entities to navigate the object hierarchy from a static origin
    """

    def __init__(
            self, game: "Game",
            parent: Optional["Renderable"] = None, surface: Optional[Surface] = None,
            position: Optional[Tuple[int, int]] = None, priority: Any = None
    ):
        Extendable.__init__(self)
        Recurface.__init__(self, surface=surface, parent=parent, position=position, priority=priority)

        self._game = game

    @property
    def game(self):
        return self._game

    def update(self, elapsed_ms: int, events: list):
        self._update(elapsed_ms, events)

        child: Renderable
        for child in reversed(self.ordered_child_recurfaces):  # Events passed to high render_priority children first
            child.update(elapsed_ms, events)

    def _update(self, elapsed_ms: int, events: list):
        """
        Lifecycle method, called automatically each game tick.
        Can optionally be overridden.
        Complete any miscellaneous work necessary for this object each tick
        that is not covered by other lifecycle methods here
        """

        pass



import os
os.environ['PYGAME_HIDE_SUPPORT_PROMPT'] = "hide"
import pygame
from pygame import Surface

from typing import Type, Optional

from .constants import Constants
from .handlers import ObserverHandler
from .screens import Screen


class Game:
    """
    Entry point for your game.
    Requires a pygame display Surface to be passed into the constructor,
    and .screen to be set to a valid Screen object, before calling .start() to begin the game loop
    """

    def __init__(self, window: Surface, fps: int = 0, updates_per_frame: int = 1):
        pygame.mixer.pre_init(frequency=Constants.AUDIO_FREQUENCY)
        pygame.init()
        pygame.mixer.init()

        self._window = window

        self.fps = fps
        self.updates_per_frame = updates_per_frame

        self._clock = pygame.time.Clock()

        self._screen = None
        self._observer_handler = ObserverHandler

    @property
    def screen(self) -> Optional[Screen]:
        return self._screen

    @screen.setter
    def screen(self, value: Screen):
        old_screen = self._screen
        self._screen = value

        self._observer_handler.on_change_screen(old_screen, value)

    @property
    def window(self):
        return self._window

    @property
    def observer_handler(self) -> Type[ObserverHandler]:
        return self._observer_handler

    def start(self) -> None:
        if self._screen is None:
            raise RuntimeError("a valid Screen object must be set to .screen before the game can be started")

        while True:
            for update_index in range(self.updates_per_frame):
                self._update_screen()

            self._render_screen()

    def _update_screen(self) -> None:
        elapsed_ms = self._clock.tick(self.fps * self.updates_per_frame)
        events = list(pygame.event.get())  # Events are passed in a list so that they can be consumed if necessary

        self._screen.update(elapsed_ms, events)

    def _render_screen(self) -> None:
        updated_rects = self._screen.render(self._window)

        pygame.display.update(updated_rects)



from typing import Any, Optional
from datetime import timedelta

from .animation import Animation
from ..handlers.enums import AnimationDataKey


class RepeatAnimation(Animation):
    def __init__(
            self, parent: "Renderable.with_extensions(Animated)", animation_key: str,
            size: float = 1, speed: float = 1, priority: Any = None,
            frame_duration: Optional[timedelta] = None,
            windup_frames: int = 0
    ):
        super().__init__(parent, animation_key, size=size, speed=speed, priority=priority)

        # Windup frames are optional non-repeating frames at the start of the animation
        self._windup_frames = windup_frames

        if frame_duration is not None:
            self._frame_time = frame_duration
        else:
            frame_duration_ms = self._settings.get(AnimationDataKey.FRAME_DURATION_MS, None)

            if frame_duration_ms is not None:
                self._frame_time = timedelta(microseconds=frame_duration_ms*1000)
            else:
                self._frame_time = Animation.default_frame_duration

    @property
    def windup_frames(self) -> int:
        return self._windup_frames

    @property
    def frame_index(self):
        frames_elapsed = int(self._elapsed_effective / self._frame_time)

        if frames_elapsed < self._windup_frames:
            return frames_elapsed
        else:
            return (
                           (frames_elapsed - self._windup_frames) %
                           (self.total_frames - self._windup_frames)
                   ) + self._windup_frames



from pygame import transform, image
from managedstate import State
from managedstate.extensions import Registrar

from typing import Iterable, Type
from os import path

from .entities import *
from .methods import Methods
from .renderable import Renderable
from .constants import Constants


class Room(Renderable):
    """
    Concrete class tightly coupled to the World screen, which renders a single room and its occupants
    """

    def __init__(self, parent: "World", room_id: str):
        super().__init__(parent.game, parent=parent, position=(0, 0), priority=0)

        self._room_id = room_id

        self._load_surface()
        self._load_room()

    @property
    def state(self) -> State.with_extensions(Registrar):
        """
        Shortcut property which accesses the state attached to the parent World object
        """

        return self.parent_recurface.state

    @property
    def room_id(self) -> str:
        return self._room_id

    def _load_surface(self, size: float = 1):
        """
        Loads the background image for the room object. Assumes a standard location for the image file as
        dictated below in `background_file_path`
        """

        background_id = self.state.registered_get("room_background_id", [self._room_id])
        background_file_path = path.join(
            Constants.RESOURCE_FOLDER_PATH,
            f"{type(self).__name__}",
            f"{background_id}.png"
        )

        surface = image.load(background_file_path).convert_alpha()
        surface = transform.rotozoom(surface, 0, size)
        self.surface = surface

    def _load_room(self):
        """
        Instantiates all the entities present in the current room.
        Assumes that any Entity subclasses listed in the game's state
        are available under the same name in your global namespace
        """

        curr_room_entities_ids: Iterable[str] = self.state.registered_get("room_entities_ids", [self._room_id])

        for entity_id in curr_room_entities_ids:
            entity_data: dict = self.state.registered_get("entity", [entity_id])

            entity_class: Type[Entity] = Methods.get_obj_by_str_name(entity_data["class"])
            entity_details: dict = entity_data["details"]

            entity_class(self, **entity_details)



from managedstate import State
from managedstate.extensions import Registrar
from managedstate.extensions.registrar import PartialQueries

from .screen import Screen
from ..room import Room


class World(Screen):
    """
    A concrete implementation of Screen, that represents a standard room-based game world
    """

    def __init__(self, game, state: State.with_extensions(Registrar)):
        super().__init__(game, state)

        self._curr_room = None
        self.set_room()

        ##### TODO: Render UI layer after room

    @property
    def curr_room(self) -> Room:
        return self._curr_room

    def set_room(self) -> None:
        new_room_id = self.state.registered_get("current_room_id")

        if self.curr_room is None:
            pass
        elif new_room_id == self.curr_room.room_id:
            return

        old_room = self._curr_room
        self._curr_room = Room(self, new_room_id)

        old_room.parent_recurface = None
        self.game.observer_handler.on_change_room(old_room, self.curr_room)

    @staticmethod
    def register_paths(state: State.with_extensions(Registrar)):
        state.register_path("current_room_id", ["current_room_id"], [str(None)])

        state.register_path("entity", ["entities", PartialQueries.KEY], [{}, {}])

        state.register_path("room", ["rooms", PartialQueries.KEY], [{}, {}])
        state.register_path("room_entities_ids", ["rooms", PartialQueries.KEY, "entities_ids"], [{}, {}, []])
        state.register_path(
            "room_background_id",
            ["rooms", PartialQueries.KEY, "background_id"],
            [{}, {}, str(None)]
        )
